/*
 * who.c -- The WHO queue.  The ISON queue.  The USERHOST queue.
 *
 * Written by Jeremy Nelson
 * Copyright 1996, 1997 EPIC Software Labs
 */

#include "irc.h"
#include "edit.h"
#include "ircaux.h"
#include "who.h"
#include "server.h"
#include "window.h"
#include "vars.h"
#include "hook.h"
#include "output.h"
#include "numbers.h"
#include "parse.h"

/*
 *
 *
 *
 * 				WHO QUEUE
 *
 *
 *
 */
static void who_queue_add (WhoEntry *item)
{
	WhoEntry *bottom = server_list[from_server].who_queue;

	while (bottom && bottom->next)
		bottom = bottom->next;

	if (!bottom)
		server_list[from_server].who_queue = item;
	else
		bottom->next = item;

	return;
}

static void who_queue_pop (void)
{
	WhoEntry *save = server_list[from_server].who_queue;

	if (save)
	{
		server_list[from_server].who_queue = save->next;
		new_free(&save->who_name);
		new_free(&save->who_host);
		new_free(&save->who_server);
		new_free(&save->who_nick);
		new_free(&save->who_real);
		new_free(&save->who_stuff);
		new_free(&save->who_end);
		new_free((char **)&save);
	}
	return;
}

static WhoEntry *who_queue_top (void)
{
	return server_list[from_server].who_queue;
}

static WhoEntry *get_new_whois_entry (void)
{
	WhoEntry *new = (WhoEntry *)new_malloc(sizeof(WhoEntry));
	new->dirty = 0;
	new->who_mask = 0;
	new->who_host = NULL;
	new->who_name = NULL;
	new->who_server = NULL;
	new->who_nick = NULL;
	new->who_real = NULL;
	new->who_stuff = NULL;
	new->who_end = NULL;
	new->next = NULL;
	who_queue_add(new);
	return new;
}


/*
 * who: the /WHO command. Parses the who switches and sets the who_mask and
 * whoo_stuff accordingly.  Who_mask and who_stuff are used in whoreply() in
 * parse.c 
 */
void whocmd (char *command, char *args, char *subargs)
{
	whobase(args, NULL, NULL);
}


/*
 * whobase: What does all the work.
 */
void whobase(char *args, void (*line) (char *, char **), void (*end) (char *, char **))
{
	char	*arg,
		*channel = NULL;
	int	no_args = 1,
		len;
	WhoEntry *new = get_new_whois_entry();

	new->line = line;
	new->end = end;

	while ((arg = next_arg(args, &args)) != NULL)
	{
		lower(arg);
		no_args = 0;

		if (*arg == '-' || *arg == '/')
		{
			arg++;
			if ((len = strlen(arg)) == 0)
			{
				say("Unknown or missing flag");
				return;
			}

			if (!strncmp(arg, "line", 4))		/* LINE */
			{
				char *line;

				if ((line = next_expr(&args, '{')))
					malloc_strcpy(&new->who_stuff, line);
				else
					say("Need {...} argument for -LINE argument.");
			}
			else if (!strncmp(arg, "end", 3))	/* END */
			{
				char *line;

				if ((line = next_expr(&args, '{')))
					malloc_strcpy(&new->who_end, line);
				else
					say("Need {...} argument for -END argument.");
			}
			else if (!strncmp(arg, "o", 1))		/* OPS */
				new->who_mask |= WHO_OPS;
			else if (!strncmp(arg, "lu", 2))	/* LUSERS */
				new->who_mask |= WHO_LUSERS;
			else if (!strncmp(arg, "ch", 2))	/* CHOPS */
				new->who_mask |= WHO_CHOPS;
			else if (!strncmp(arg, "ho", 2))	/* HOSTS */
			{
				if ((arg = next_arg(args, &args)) == NULL)
				{
					say("WHO -HOST: missing argument");
					return;
				}

				new->who_mask |= WHO_HOST;
				malloc_strcpy(&new->who_host, arg);
				channel = new->who_host;
			}
		 	else if (!strncmp(arg, "he", 2))	/* here */
				new->who_mask |= WHO_HERE;
			else if (!strncmp(arg, "a", 1))		/* away */
				new->who_mask |= WHO_AWAY;
			else if (!strncmp(arg, "s", 1)) 	/* servers */
			{
				if ((arg = next_arg(args, &args)) == NULL)
				{
					say("WHO -SERVER: missing arguement");
					return;
				}

				new->who_mask |= WHO_SERVER;
				malloc_strcpy(&new->who_server, arg);
				channel = new->who_server;
			}
			else if (!strncmp(arg, "na", 2))
			{
				if ((arg = next_arg(args, &args)) == NULL)
				{
					say("WHO -NAME: missing arguement");
					return;
				}

				new->who_mask |= WHO_NAME;
				malloc_strcpy(&new->who_name, arg);
				channel = new->who_name;
			}
			else if (!strncmp(arg, "r", 1))
			{
				if ((arg = next_arg(args, &args)) == NULL)
				{
					say("WHO -REALNAME: missing arguement");
					return;
				}

				new->who_mask |= WHO_REAL;
				malloc_strcpy(&new->who_real, arg);
				channel = new->who_real;
			}
			else if (!strncmp(arg, "ni", 2))
			{
				if ((arg = next_arg(args, &args)) == NULL)
				{
					say("WHO -NICK: missing arguement");
					return;
				}

				new->who_mask |= WHO_NICK;
				malloc_strcpy(&new->who_nick, arg);
				channel = new->who_nick;
			}
			else
			{
				say("Unknown or missing flag");
				return;
			}
		}
		else if (strcmp(arg, "*") == 0)
		{
			channel = get_channel_by_refnum(0);
			if (!channel || *channel == '0')
			{
				say("I wouldn't do that if I were you");
				return;
			}
		}
		else
			channel = arg;
	}

	if (no_args)
		say("No argument specified");
	else
	{
		if (!channel && (new->who_mask & WHO_OPS))
			channel = "*.*";
		send_to_server("WHO %s %c", channel ? channel : empty_string, 
				(new->who_mask & WHO_OPS) ?  'o' : '\0');
	}
}


void whoreply (char *from, char **ArgList)
{
static	char	format[40];
static	int	last_width = -1;
	int	ok = 1;
	char	*channel,
		*user,
		*host,
		*server,
		*nick,
		*stat,
		*name;
	WhoEntry *new = who_queue_top();

	if (!ArgList[5])
		return;		/* Fake! */

	if (!new)
	{
		yell("### Please dont do /QUOTE WHO.");
		return;
	}

	/*
	 * We dont always want to use this function.
	 * If another function is supposed to do the work for us,
	 * we yield to them.
	 */
	if (new->line)
	{
		new->line(from, ArgList);
		return;
	}

	if (last_width != get_int_var(CHANNEL_NAME_WIDTH_VAR))
	{
		if ((last_width = get_int_var(CHANNEL_NAME_WIDTH_VAR)) != 0)
		    sprintf(format, "%%-%u.%us %%-9s %%-3s %%s@%%s (%%s)",
					(unsigned) last_width,
					(unsigned) last_width);
		else
		    strcpy(format, "%s\t%-9s %-3s %s@%s (%s)");
	}
	channel = ArgList[0];
	user    = ArgList[1];
	host    = ArgList[2];
	server  = ArgList[3];
	nick    = ArgList[4];
	stat    = ArgList[5];
	PasteArgs(ArgList, 6);
	name    = ArgList[6];

	if (*stat == 'S')	/* this only true for the header WHOREPLY */
	{
		char buffer[1024];

		channel = "Channel";
		snprintf(buffer, 1024, "%s %s %s %s %s %s %s", channel,
				nick, stat, user, host, server, name);

		if (new->who_stuff)
			;			/* munch it */

		else if (do_hook(WHO_LIST, "%s", buffer))
			put_it(format, channel, nick, stat, user, host, name);
		return;
	}

	if (new && new->who_mask)
	{
		if (new->who_mask & WHO_HERE)
			ok = ok && (*stat == 'H');
		if (new->who_mask & WHO_AWAY)
			ok = ok && (*stat == 'G');
		if (new->who_mask & WHO_OPS)
			ok = ok && (*(stat + 1) == '*');
		if (new->who_mask & WHO_LUSERS)
			ok = ok && (*(stat + 1) != '*');
		if (new->who_mask & WHO_CHOPS)
			ok = ok && ((*(stat + 1) == '@') ||
				    (*(stat + 2) == '@'));
		if (new->who_mask & WHO_NAME)
			ok = ok && wild_match(new->who_name, user);
		if (new->who_mask & WHO_NICK)
			ok = ok && wild_match(new->who_nick, nick);
		if (new->who_mask & WHO_HOST)
			ok = ok && wild_match(new->who_host, host);
		if (new->who_mask & WHO_REAL)
			ok = ok && wild_match(new->who_real, name);
		if (new->who_mask & WHO_SERVER)
			ok = ok && wild_match(new->who_server, server);
	}

	if (ok)
	{
		char buffer[1024];

		snprintf(buffer, 1023, "%s %s %s %s %s %s %s", channel,
				nick, stat, user, host, server, name);

		if (new->who_stuff)
			parse_line(NULL, new->who_stuff, buffer, 0, 0);

		else if (do_hook(WHO_LIST, "%s", buffer))
		{
			if (!get_int_var(SHOW_WHO_HOPCOUNT_VAR))
				next_arg(name, &name);

			put_it(format, channel, nick, stat, user, host, name);
		}
	}
}


void who_end (char *from, char **ArgList)
{
	WhoEntry *new = who_queue_top();
	char buffer[1024];

	if (!new)
		return;		/* Lame user. */

	/*
	 * Defer to another function, if neccesary.
	 */
	if (new->end)
	{
		new->end(from, ArgList);
		return;
	}

	snprintf(buffer, 1024, "%s %s", from, ArgList[0]);
	if (new->who_end)
		parse_line(NULL, new->who_end, buffer, 0, 0);

	else if (get_int_var(SHOW_END_OF_MSGS_VAR))
		if (do_hook(current_numeric, "%s", buffer))
			put_it("%s %s", numeric_banner(), buffer);

	who_queue_pop();
}


/*
 *
 *
 *
 * 				ISON QUEUE
 *
 *
 *
 */
static void ison_queue_add (IsonEntry *item)
{
	IsonEntry *bottom = server_list[from_server].ison_queue;

	while (bottom && bottom->next)
		bottom = bottom->next;

	if (!bottom)
		server_list[from_server].ison_queue = item;
	else
		bottom->next = item;

	return;
}

static void ison_queue_pop (void)
{
	IsonEntry *save = server_list[from_server].ison_queue;
	if (save)
	{
		server_list[from_server].ison_queue = save->next;
		new_free(&save->ison_asked);
		new_free(&save->ison_got);
		new_free((char **)&save);
	}
	return;
}

static IsonEntry *ison_queue_top (void)
{
	return server_list[from_server].ison_queue;
}

static IsonEntry *get_new_ison_entry (void)
{
	IsonEntry *new = (IsonEntry *)new_malloc(sizeof(IsonEntry));
	new->ison_asked = NULL;
	new->ison_got = NULL;
	new->next = NULL;
	new->line = NULL;
	ison_queue_add(new);
	return new;
}


void isoncmd (char *command, char *args, char *subargs)
{
	if (!args || !*args)
		args = get_server_nickname(from_server);

	isonbase(args, NULL);
}

void isonbase (char *args, void (*line) (char *, char *))
{
	IsonEntry *new = get_new_ison_entry();

	new->line = line;
	malloc_strcpy(&new->ison_asked, args);
	send_to_server("ISON %s", new->ison_asked);
}

/* 
 * ison_returned: this is called when numeric 303 is received in
 * numbers.c. ISON must always be the property of the WHOIS queue.
 * Although we will check first that the top element expected is
 * actually an ISON.
 */
void	ison_returned (char *from, char **ArgList)
{
	IsonEntry *new = ison_queue_top();

	if (!new)
	{
		yell("### Please dont do /QUOTE ISON.");
		return;
	}

	PasteArgs(ArgList, 0);
	if (new->line)
		new->line(new->ison_asked, ArgList[0]);
	else
	{
		if (do_hook(current_numeric, "%s", ArgList[0]))
			put_it("%s Currently online: %s", numeric_banner(), ArgList[0]);
	}

	ison_queue_pop();
	return;
}


#if 0
/*
 *
 *
 *
 *
 *				USERHOST QUEUE
 *
 *
 *
 *
 */
static void userhost_queue_add (UserhostEntry *item)
{
	UserhostEntry *bottom = server_list[from_server].userhost_queue;

	while (bottom && bottom->next)
		bottom = bottom->next;

	if (!bottom)
		server_list[from_server].userhost_queue = item;
	else
		bottom->next = item;

	return;
}

static void userhost_queue_pop (void)
{
	UserhostEntry *save = server_list[from_server].userhost_queue;
	server_list[from_server].userhost_queue = save->next;

	new_free(&save->userhost_asked);
	new_free(&save->userhost_got);
	new_free((char **)&save);
	return;
}

static UserhostEntry *userhost_queue_top (void)
{
	return server_list[from_server].userhost_queue;
}

static UserhostEntry *get_new_userhost_entry (void)
{
	UserhostEntry *new = (UserhostEntry *)new_malloc(sizeof(UserhostEntry));
	new->userhost_asked = NULL;
	new->userhost_got = NULL;
	new->next = NULL;
	new->line = NULL;
	new->func = NULL;
	userhost_queue_add(new);
	return new;
}

/*
 * userhost: Does the USERHOST command.  Need to split up the queries,
 * since the server will only reply to 5 at a time.
 */
BUILT_IN_COMMAND(userhost)
{
	userhostbase(args, NULL);
}

void userhostbase(char *arg, void (*line) (char *, char *))
{
	int	n = 0,
		total = 0,
		userhost_cmd = 0;
	char	*nick;
	char	buffer[BIG_BUFFER_SIZE + 1];
	char 	*ptr, 
		*next_ptr,
		*body;

	*buffer = 0;

	while ((nick = next_arg(args, &args)) != NULL)
	{
		if (check_nickname(nick))
		{
			if (n++)
				strmcat(buffer, " ", BIG_BUFFER_SIZE);

			strmcat(buffer, nick, BIG_BUFFER_SIZE);
		}

		else if (!my_strnicmp(nick, "-cmd", 2))
		{
			userhost_cmd = 1;
			while (my_isspace(*args))
				args++;

			if ((body = next_expr_failok(&args, '{')))
				args = body;

			break;
		}
	}

	if (!n)
		strmcpy(buffer, get_server_nickname(from_server));

	ptr = buffer;

	while (ptr && *ptr)
	{
		ptr = buffer;
		move_to_abs_word(ptr, &next_ptr, 5);

		if (next_ptr && *next_ptr && next_ptr > ptr)
			next_ptr[-1] = 0;

		UserhostEntry *new = get_new_userhost_entry();
		new->userhosts_asked = m_strdup(ptr);
		send_to_server("USERHOST %s", new->userhosts_asked);

		if (userhost_cmd)
			new->line = m_strdup(args);
		else
			new->func = userhost_returned;

		ptr = next_ptr;
	}
}

/* 
 * userhost_returned: this is called when numeric 302 is received in
 * numbers.c. USERHOST must also remain the property of the WHOIS
 * queue. Sending it without going via the WHOIS queue will cause
 * the queue to become corrupted.
 *
 * While USERHOST gives us similar information to WHOIS, this routine's
 * format is a little different to those that handle the WHOIS numerics.
 * A list of nicks can be supplied to USERHOST, and any of those
 * nicks are not currently signed on, it will just omit reporting it.
 * this means we must go through the list of nicks returned and check
 * them for missing entries. This means stepping through the requested
 * nicks one at a time.
 *
 * A side effect of this is that user initiated USERHOST requests get
 * reported as separate replies even though one request gets made and
 * only one reply is actually received. This should make it easier for
 * people wanting to use USERHOST for their own purposes.
 *
 * In this routine, cnick points to all the information in the next
 * entry from the returned data.
 */
void	userhost_returned (char *from, char **ArgList)
{
	/*
	 * Get the userhost queue top.
	 * Each item in ``ArgList'' is a single return value from
	 * a valid nickname.  Absent nicknames are silently ignored,
	 * so we have to check for that.  Get an arg, strip it apart
	 * check it against the first item in the list, send it off to
	 * the user.
	 */


	char	*nick,
		*cnick = NULL,
		*tnick;
	char	*user,
		*cuser = (char *) 0;
	char	*host,
		*chost = (char *) 0;
	int	ishere,
		cishere = 1;
	int	isoper,
		cisoper = 0;
	int	noton,
		isuser,
		parsed;
	char	*queue_nicks;

	if (!ArgList[0])
		return;

	parsed = 0;
	while ((parsed || (cnick = next_arg(ArgList[0], ArgList))) || queue_nicks)
	{
		if (queue_nicks)
		{
			tnick = next_arg(queue_nicks, &queue_nicks);
			if (!*queue_nicks)
				queue_nicks = (char *) 0;
		}
		else
			tnick = NULL;

		if (cnick && !parsed)
		{
			if (!(cuser = index(cnick,'=')))
				break;
			if (cuser[-1] == '*')
			{
				cuser[-1] = 0;
				cisoper = 1;
			}
			else
				cisoper = 0;

			*cuser++ = 0;
			if (*cuser++ == '+')
				cishere = 1;
			else
				cishere = 0;

			if (!(chost = index(cuser, '@')))
				break;
			*chost++ = 0;
			parsed = 1;
		}

		if (!cnick || (tnick && my_stricmp(cnick, tnick)))
		{
			if (tnick)
				nick = tnick;
			else
				nick = cnick;
			user = host = "<UNKNOWN>";
			isoper = 0;
			ishere = 1;
			noton = 1;
		}
		else
		{
			nick = cnick;
			user = cuser;
			host = chost;
			isoper = cisoper;
			ishere = cishere;
			noton = parsed = 0;
		}

		if (!isuser)
		{
			malloc_strcpy(&whois_stuff->nick, nick);
			malloc_strcpy(&whois_stuff->user, user);
			malloc_strcpy(&whois_stuff->host, host);
			whois_stuff->oper = isoper;
			whois_stuff->not_on = noton;
			if (!ishere)
				malloc_strcpy(&whois_stuff->away, empty_string);
			else
				new_free(&whois_stuff->away);
			thing->func(whois_stuff, tnick, thing->text);
			new_free(&whois_stuff->away);
		}
		else
		{
			if (do_hook(current_numeric, "%s %s %s %s %s", nick,
			    isoper ? "+" : "-", ishere ? "+" : "-", user, host))
				put_it("%s %s is %s@%s%s%s", numeric_banner(),
					nick, user, host, isoper ?
					" (Is an IRC operator)" : empty_string,
					ishere ? empty_string : " (away)");
		}
	}
	if (thing)
	{
		new_free(&thing->nick);
		new_free(&thing->text);
		new_free((char **)&thing);
	}
}

void	userhost_cmd_returned (WhoisStuff *stuff, char *nick, char *text)
{
	char	args[BIG_BUFFER_SIZE + 1];

	strcpy(args, stuff->nick ? stuff->nick : empty_string);
	strcat(args, stuff->oper ? " + " : " - ");
	strcat(args, stuff->away ? "+ " : "- ");
	strcat(args, stuff->user ? stuff->user : empty_string);
	strcat(args, space);
	strcat(args, stuff->host ? stuff->host : empty_string);
	parse_line((char *) 0, text, args, 0, 0);
}

#endif
